@@ -59,6 +59,10 @@
 :- use_module(library(prolog_history), [prolog_history/1]).
 :- use_module(library(swi_preferences), [prolog_edit_preferences/1]).
 :- use_module(library(pce_openframes), [confirm_open_frames/1]).
+:- use_module(library(ansi_term), [ansi_format/3]).
+:- use_module(library(error), [existence_error/2]).
+:- use_module(library(prolog_code), [pi_head/2]).
+:- use_module(library(thread), [call_in_thread/2, call_in_thread/3]).
 
 :- meta_predicate
     epilog(:),
@@ -66,8 +70,6 @@
     run_in_help_epilog(0),
     win_insert_menu_item(+, +, +, 0).
 
-:- pce_autoload(finder, library(find_file)).
-:- pce_global(@finder, new(finder)).
 :- pce_global(@epilog, new(epilog)).
 
 
@@ -187,7 +189,7 @@ epilog_attach(Options) :-
     set_prolog_flag(save_history, false),
     in_pce_thread(create_epilog(TID, Options)),
     thread_get_message('$epilog'(PT, PTY)),
-    thread_at_exit(terminated),
+    prolog_listen(this_thread_exit, terminated),
     set_prolog_flag(query_debug_settings, debug(false, false)),
     set_prolog_flag(hyperlink_term, true),
     attach_terminal(PT, PTY, _Title, []),
@@ -292,6 +294,10 @@ binding('\\C-\\S-w', close).
 binding('\\C-\\S-=', font_magnify).
 binding('\\C--',     font_reduce).
 binding('\\C-=',     font_default).
+binding('<f5>',      trace_mode).
+binding('\\S-<f5>',  debug_mode).
+binding('\\C-<f5>',  gui_debug).
+binding('<f6>',      debugging).
 :- if(current_prolog_flag(apple, true)).
 binding(Key, Method) :-
     pce_keybinding:binding(apple, epilog, Bindings),
@@ -367,12 +373,12 @@ unlink(PT) :->
     catch(unlink_terminal_thread(PT), error(_,_), true),
     send_super(PT, unlink).
 
-unlink_terminal_thread(PT) :-
+unlink_terminal_thread(PT) :- % Epilog attached to a running thread
     retract(attached_terminal(PT, RestoreContext)),
     retract(current_prolog_terminal(Thread, PT)),
     !,
     call_in_thread(Thread, restore_io(RestoreContext)).
-unlink_terminal_thread(PT) :-
+unlink_terminal_thread(PT) :- % Normal Epilog window
     retract(current_prolog_terminal(Thread, PT)),
     !,
     thread_signal(Thread, clean_exit).
@@ -405,7 +411,7 @@ terminated :-
     delete_window,
     close_io.
 
-delete_window :-
+delete_window :-            % Epilog attached to running thread
     thread_self(Me),
     thread_property(Me, id(Id)),
     current_prolog_terminal(Me, PT),
@@ -414,11 +420,13 @@ delete_window :-
     get_time(Now),
     format_time(string(T), '%+', Now),
     ansi_format(comment, '~N<thread ~w finished at ~s>~n', [Id, T]).
-delete_window :-
+delete_window :-            % Normal Epilog
     thread_self(Me),
-    retract(current_prolog_terminal(Me, PT)),
+    current_prolog_terminal(Me, PT),
     !,
     save_history(PT),
+    retractall(terminal_input(PT, _Pty, _In, _Out, _Error, _Edit)),
+    retractall(current_prolog_terminal(Me, PT)),
     (   '$run_state'(normal)
     ->  in_pce_thread(send(PT?frame, delete_epilog, PT?window))
     ;   true
@@ -494,6 +502,45 @@ interrupt(PT) :->
     current_signal(int, SIGINT, debug),
     thread_signal(Thread, SIGINT).
 
+debug_mode(PT) :->
+    "Toggle Prolog debug mode"::
+    (   terminal_prolog_flag(PT, query_debug_settings,
+                             debug(Debugging, _Tracing), -)
+    ->  debug_toggle_command(Debugging, Negate),
+        send(PT, inject, Negate)
+    ;   true
+    ).
+
+debug_toggle_command(true, nodebug).
+debug_toggle_command(false, debug).
+
+trace_mode(PT) :->
+    "Toggle Prolog trace mode"::
+    (   terminal_prolog_flag(PT, query_debug_settings,
+                             debug(_Debugging, Tracing), -)
+    ->  trace_toggle_command(Tracing, Negate),
+        send(PT, inject, Negate)
+    ;   true
+    ).
+
+trace_toggle_command(true, notrace).
+trace_toggle_command(false, trace).
+
+debugging(PT) :->
+    "Show debugging status"::
+    send(PT, inject, debugging).
+
+gui_debug(PT) :->
+    "Toggle Prolog GUI tracer"::
+    (   terminal_prolog_flag(PT, gui_tracer, GuiDebug, false)
+    ->  gui_debug_toggle_command(GuiDebug, Negate),
+        send(PT, inject, Negate)
+    ;   true
+    ).
+
+gui_debug_toggle_command(true,  noguitracer).
+gui_debug_toggle_command(false, guitracer).
+
 close(PT) :->
     "Close this Prolog shell"::
     get(PT, window, Window),
@@ -502,11 +549,19 @@ close(PT) :->
     get_time(Now),
     asserta(closed_epilog(Epilog, Now)).
 
+%!  save_history(+PrologTerminal) is det.
+%
+%   Save the history for PrologTerminal.
+
 save_history(PT) :-
-    retract(terminal_input(PT, _PTY, _In, _Out, _Err, EditLine)),
-    EditLine == true,
+    terminal_input(PT, _PTY, _In, _Out, _Err, true),
+    current_prolog_terminal(Thread, PT),
     !,
-    prolog_history(save).               % If loaded and `save_history` is `true`
+    call_in_thread(Thread,
+                   catch(prolog_history(save), error(_,_), true),
+                   [ timeout(0.1),
+                     on_timeout(true)
+                   ]).
 save_history(_).
 
 history_events(PT, Events:prolog) :<-
@@ -538,15 +593,13 @@ history_events(_PT, []) :-
     !.
 history_events(_PT, load) :-
     !,
-    prolog_history(enable),
-    '$load_history'.
+    prolog_history(enable).
 history_events(PT, Events) :-
     terminal_input(PT, _PTY, In, _Out, _Err, true),
     stream_property(In, file_no(Fd)),
     reverse(Events, OldFirst),
     forall(member(_N-Line, OldFirst),
-           el_add_history(Fd, Line)),
-    editline:load_history_events(Events).
+           el_add_history(Fd, Line)).
 
 history(PT, Enabled:enable={on,off,copy}, Save:save=[bool]) :->
     "Enable/disable history"::
@@ -564,7 +617,6 @@ activate_history(PT) :-
     ->  prolog_history(disable)
     ;   Enabled == on
     ->  prolog_history(enable),
-        '$load_history',
         (   Save \== @on
         ->  set_prolog_flag(save_history, false)
         ;   true
@@ -731,7 +783,7 @@ thread_run_interactor(PT, Creator, PTY, Init, Goal, Title, History) :-
     ->  (   var(Formal)
         ->  thread_send_message(Creator, title(Title)),
             call(Init),
-            call(Goal)
+            ignore(Goal)
         ;   thread_send_message(Creator, throw(Error))
         )
     ;   thread_send_message(Creator, false)
@@ -859,6 +911,11 @@ split(T, Dir:{horizontally,vertically}) :->
     ;   send(W, right, Tile)
     ).
 
+save_history(EW) :->
+    "Save the commandline history"::
+    get(EW, terminal, PT),
+    save_history(PT).
+
 :- pce_end_class(epilog_window).
 
 
@@ -898,6 +955,10 @@ epilog_name(@default, _, Name) :-
     gensym(epilog, Name).
 epilog_name(Name, _, Name).
 
+destroy(Epilog) :->
+    "Destroy the Epilog terminal"::
+    send(Epilog, save_history),
+    send_super(Epilog, destroy).
 
 % ->wm_close_requested
 %
@@ -921,7 +982,8 @@ delete_epilog(T, W:window, Destroy:[bool]) :->
             Terminals),
         send(Terminals, delete, W),
         \+ send(Terminals, empty)
-    ->  send(T, delete, W),
+    ->  send(W, save_history),
+        send(T, delete, W),
         (   Destroy == @on
         ->  send(W, destroy)
         ;   true
@@ -967,37 +1029,77 @@ inject(Epilog, Command:prolog) :->
 
 consult(T) :->
     "Ask for a file and consult it"::
-    findall(Ext, user:prolog_file_type(Ext, prolog), Exts),
-    chain_list(Filter, Exts),
-    get(@finder, file, open, tuple('Prolog file', Filter), File),
-    send(T, inject, consult(File)).
+    source_file_filter(Filter),
+    working_directory(CWD, CWD),
+    get(T?frame, open_file,
+        filters := Filter,
+        default := CWD,
+        allow_many := @on, FileChain),
+    chain_list(FileChain, Files),
+    send(T, inject, consult(Files)).
 
-edit_file(_T) :->
+edit_file(T) :->
     "Ask for a file and edit it"::
-    findall(Ext, user:prolog_file_type(Ext, source), Exts),
-    chain_list(Filter, Exts),
-    get(@finder, file, open, tuple('Source', Filter), File),
+    source_file_filter(Filter),
+    (   current_prolog_flag(associated_file, Default)
+    ->  true
+    ;   working_directory(Default, Default)
+    ),
+    get(T?frame, open_file,
+        filters := Filter,
+        default := Default,
+        File),
     edit(file(File)).
 
-new_file(_T) :->
+new_file(T) :->
     "Ask for a file and create it"::
-    findall(Ext, user:prolog_file_type(Ext, source), Exts),
-    chain_list(Filter, Exts),
-    get(@finder, file, save, tuple('Source', Filter), File),
+    source_file_filter(Filter),
+    working_directory(CWD, CWD),
+    get(T?frame, save_file,
+        filters := Filter,
+        default := CWD,
+        File0),
+    ensure_prolog_extension(File0, File),
     edit(file(File)).
 
+source_file_filter(Filter) :-
+    findall(Ext, user:prolog_file_type(Ext, source), Exts),
+    chain_list(ExtChain, Exts),
+    new(Filter, chain(tuple('Source', ExtChain))).
+
+%!  ensure_prolog_extension(+File0, -File) is det.
+%
+%   Ensure File has a Prolog extension.
+
+ensure_prolog_extension(File0, File) :-
+    file_name_extension(_, Ext, File0),
+    user:prolog_file_type(Ext, prolog),
+    !,
+    File = File0.
+ensure_prolog_extension(File0, File) :-
+    file_name_extension(File0, pl, File).
+
 make(T) :->
     "Run make/0"::
     send(T, inject, make).
 
-quit(T, Prolog:prolog=[bool]) :->
-    "Quit this terminal.  Optionally should terminate Prolog"::
+close(T, Prolog:prolog=[bool]) :->
+    "Close this terminal.  Optionally terminates Prolog"::
     send(T, destroy),
     (   Prolog == @on
     ->  assert(quit_requested)
     ;   true
     ).
 
+save_history(Epilog) :->
+    "Save pending history"::
+    (   terminal_input(PT, _PTY, _In, _Out, _Err, true),
+        get(PT, frame, Epilog),
+        save_history(PT),
+        fail
+    ;   true
+    ).
+
 interrupt(Epilog) :->
     "Interrupt running thread"::
     get(Epilog, current_terminal, Term),
@@ -1015,6 +1117,75 @@ open_url(_T, URL:name) :->
     "Open a URL"::
     www_open_url(URL).
 
+debug_mode(Frame) :->
+    "Toggle Prolog debug mode"::
+    get(Frame, current_terminal, Term),
+    send(Term, debug_mode).
+
+trace_mode(Frame) :->
+    "Toggle Prolog trace mode"::
+    get(Frame, current_terminal, Term),
+    send(Term, trace_mode).
+
+debugging(Frame, MI:menu_item) :->
+    "Toggle Prolog debug mode"::
+    get(Frame, current_terminal, Term),
+    send(Term, debugging),
+    send(MI, selected, @off).
+
+gui_debug(Frame) :->
+    "Toggle GUI tracer"::
+    get(Frame, current_terminal, Term),
+    send(Term, gui_debug).
+
+update_debug_mode(Frame, MI:menu_item) :->
+    "Updated the current debug mode"::
+    (   epilog_prolog_flag(Frame, query_debug_settings,
+                           debug(Debugging, _Tracing), -)
+    ->  send(MI, selected, Debugging)
+    ;   true
+    ).
+
+update_trace_mode(Frame, MI:menu_item) :->
+    "Updated the current trace mode"::
+    (   epilog_prolog_flag(Frame, query_debug_settings,
+                           debug(_Debugging, Tracing), -)
+    ->  send(MI, selected, Tracing)
+    ;   true
+    ).
+
+update_gui_debug(Frame, MI:menu_item) :->
+    (   epilog_prolog_flag(Frame, gui_tracer, GuiTracer, false)
+    ->  send(MI, selected, GuiTracer)
+    ;   true
+    ).
+
+epilog_prolog_flag(Frame, Flag, Value, Default) :-
+    get(Frame, current_terminal, Term),
+    terminal_prolog_flag(Term, Flag, Value, Default).
+
+%!  terminal_prolog_flag(+Term, +Flag, -Value, +Default) is semidet.
+%
+%   Get the Prolog flat Flag for the toplevel thread running in Term. If
+%   the flag is not defined, unify   Value  with Default. This predicate
+%   uses a timeout of 0.1 seconds,   returning  Default on timeout. This
+%   guarantees that the console will not  freeze   if  the thread is not
+%   responsive.
+
+terminal_prolog_flag(Term, Flag, Value, Default) :-
+    current_prolog_terminal(Thread, Term),
+    (   catch(call_in_thread(Thread,
+                             current_prolog_flag(Flag, Value),
+                             [ timeout(0.1),
+                               on_timeout(fail)
+                             ]),
+              error(Formal,_),
+              true)
+    ->  var(Formal)
+    ;   Value = Default
+    ).
+
+
 :- pce_end_class(epilog_frame).
 
 :- pce_begin_class(epilog_dialog, dialog, "Prolog terminator menu").
@@ -1027,6 +1198,7 @@ initialise(D) :->
     send(MB, append, new(File,     popup(file))),
     send(MB, append, new(Settings, popup(settings))),
     send(MB, append, new(Tools,    popup(tools))),
+    send(MB, append, new(Debug,    epilog_popup(debug))),
     send(MB, append, new(Help,     popup(help))),
     Epilog = @event?receiver?frame,
     send_list(File, append,
@@ -1040,10 +1212,10 @@ initialise(D) :->
                 menu_item(reload_modified_files,
                           message(Epilog, make),
                           end_group := @on),
-                menu_item(quit,
-                          message(Epilog, quit)),
-                menu_item(quit_prolog,
-                          message(Epilog, quit, @on))
+                menu_item(close,
+                          message(Epilog, close)),
+                menu_item(halt_prolog,
+                          message(Epilog, close, @on))
               ]),
     send_list(Settings, append,
               [ menu_item(user_init_file,
@@ -1059,7 +1231,29 @@ initialise(D) :->
                 menu_item(debug_messages,
                           message(Epilog, ide, debug_monitor)),
                 menu_item(cross_referencer,
-                          message(Epilog, ide, xref))
+                          message(Epilog, ide, xref),
+                          end_group := @on),
+                menu_item('Inspect GUI hierarchy',
+                          message(Epilog, ide, visual_hierarchy),
+                          end_group := @on)
+              ]),
+    send_list(Debug, append,
+              [ new(TraceMode,
+                    menu_item(trace_mode,
+                              message(Epilog, trace_mode),
+                              accelerator := 'F5')),
+                new(DebugMode,
+                    menu_item(debug_mode,
+                              message(Epilog, debug_mode),
+                              accelerator := 'Shift-F5')),
+                new(GuiDebug,
+                    menu_item('GUI_debugger',
+                              message(Epilog, gui_debug),
+                              accelerator := 'Ctrl-F5',
+                              end_group := @on)),
+                menu_item(show_debug_status,
+                          message(Epilog, debugging, @arg1),
+                          accelerator := 'F6')
               ]),
     send_list(Help, append,
               [ menu_item('SWI-Prolog documentation',
@@ -1072,7 +1266,12 @@ initialise(D) :->
                 menu_item('SWI-Prolog GUI tools',
                           message(Epilog, open_url,
                                   'https://github.com/SWI-Prolog/packages-xpce/wiki'))
-              ]).
+              ]),
+    send(Debug, show_current, @on),
+    send(Debug, multiple_selection, @on),
+    send(DebugMode, condition, message(Epilog, update_debug_mode, DebugMode)),
+    send(TraceMode, condition, message(Epilog, update_trace_mode, TraceMode)),
+    send(GuiDebug,  condition, message(Epilog, update_gui_debug, GuiDebug)).
 
 :- pce_end_class(epilog_dialog).
 
@@ -1111,13 +1310,10 @@ capture_messages :-
 
 :- public xpce_message/3.
 xpce_message(_Term, Kind, Lines) :-
-    report_kind(Kind),
+    Kind \== silent,
     xpce_epilog_console(_In,_Out,Error),
     print_message_lines(Error, kind(Kind), Lines).
 
-report_kind(warning).
-report_kind(error).
-
 %!  pce:xpce_console(-In,-Out,-Error) is semidet.
 %
 %   Tell xpce where to write console output.   If  this fails, output is
